<!ENTITY asdf "<application>ASDF</application>">
<!ENTITY clocc "<application><emphasis>CLOCC</emphasis></application>">
<!ENTITY cclan "<application><emphasis>CCLAN</emphasis></application>">
<!ENTITY uffi "<application><emphasis>UFFI</emphasis></application>">
<!ENTITY ffi "<emphasis>FFI</emphasis>">
<!ENTITY clsql "<application><emphasis>CLSQL</emphasis></application>">
<!ENTITY maisql "<application><emphasis>MaiSQL</emphasis></application>">
<!ENTITY sql "<application>SQL</application>">
<!ENTITY usql "<application>UncommonSQL</application>">
<!ENTITY commonsql "<application>CommonSQL</application>">
<!ENTITY mysql "<application>MySQL</application>">
<!ENTITY postgresql "<application>PostgreSQL</application>">
<!ENTITY sqlite "<application>SQLite version 2</application>">
<!ENTITY sqlite3 "<application>SQLite version 3</application>">
<!ENTITY oracle "<application>Oracle</application>">
<!ENTITY odbc "<application>ODBC</application>">
<!ENTITY aodbc "<application>AODBC</application>">
<!ENTITY cmucl "<application>CMUCL</application>">
<!ENTITY scl "<application>SCL</application>">
<!ENTITY md5 "<application>MD5</application>">
<!ENTITY sbcl "<application>SBCL</application>">
<!ENTITY openmcl "<application>OpenMCL</application>">
<!ENTITY lw "<application>Lispworks</application>">
<!ENTITY acl "<application>AllegroCL</application>">
<!ENTITY cl "<application>ANSI Common Lisp</application>">
<!ENTITY t "<constant>T</constant>">
<!ENTITY nil "<constant>NIL</constant>">
<!ENTITY null "<constant>NULL</constant>">
<!ENTITY c "<application>C</application>">
<!ENTITY unix "<application>UNIX</application>">
<!ENTITY mswindows "<application>Microsoft Windows</application>">
